# -*- coding: utf-8 -*-

# businessdate
# ------------
# Python library for generating business dates for fast date operations
# and rich functionality.
# 
# Author:   sonntagsgesicht, based on a fork of Deutsche Postbank [pbrisk]
# Version:  0.5, copyright Wednesday, 18 September 2019
# Website:  https://github.com/sonntagsgesicht/businessdate
# License:  Apache License 2.0 (see LICENSE file)


from datetime import date, timedelta

from .ymd import from_excel_to_ymd, from_ymd_to_excel


class BaseDateFloat(float):
    """ native :class:`float` backed base class for a performing date calculations counting days since Jan, 1st 1900 """

    def __new__(cls, x=0):
        new = super(BaseDateFloat, cls).__new__(cls, x)
        new._ymd = None
        return new

    # --- property methods ---------------------------------------------------

    @property
    def day(self):
        if not self._ymd:
            self._ymd = self.to_ymd()
        return self._ymd[2]

    @property
    def month(self):
        if not self._ymd:
            self._ymd = self.to_ymd()
        return self._ymd[1]

    @property
    def year(self):
        if not self._ymd:
            self._ymd = self.to_ymd()
        return self._ymd[0]

    def weekday(self):
        return self.to_date().weekday()

    # --- constructor method -------------------------------------------------

    @classmethod
    def from_ymd(cls, year, month, day):
        """ creates instance from a :class:`tuple` of :class:`int` items `(year, month, day)` """
        return cls(from_ymd_to_excel(year, month, day))

    @classmethod
    def from_date(cls, d):
        """ creates instance from a :class:`datetime.date` object `d` """
        return cls.from_ymd(d.year, d.month, d.day)

    @classmethod
    def from_float(cls, x):
        """ creates from a :class:`float` `x` counting the days since Jan, 1st 1900 """
        return cls(x)

    # --- cast method --------------------------------------------------------

    def to_ymd(self):
        """ returns the :class:`tuple` of :class:`int` items `(year, month, day)` """
        if not self._ymd:
            self._ymd = from_excel_to_ymd(int(self))
        return self._ymd

    def to_date(self):
        """ returns `datetime.date(year, month, day)` """
        if not self._ymd:
            self._ymd = self.to_ymd()
        return date(*self._ymd)

    def to_float(self):
        """ returns :class:`float` counting the days since Jan, 1st 1900 """
        return float(self)

    # --- calculation methods ------------------------------------------------

    def _add_days(self, n):
        self._ymd = None
        return self.__class__(super(BaseDateFloat, self).__add__(n))

    def _diff_in_days(self, d):
        return super(BaseDateFloat, d).__sub__(float(self))


class BaseDateDatetimeDate(date):
    """ :class:`datetime.date` backed base class for a performing date calculations """

    # --- constructor method -------------------------------------------------

    @classmethod
    def from_ymd(cls, year, month, day):
        """ creates instance from a :class:`tuple` of :class:`int` items `(year, month, day)` """
        return cls(year, month, day)

    @classmethod
    def from_date(cls, d):
        """ creates instance from a :class:`datetime.date` object `d` """
        return cls.from_ymd(d.year, d.month, d.day)

    @classmethod
    def from_float(cls, x):
        """ creates from a :class:`float` `x` counting the days since Jan, 1st 1900 """
        y, m, d = from_excel_to_ymd(x)
        return cls.from_ymd(y, m, d)

    # --- cast method --------------------------------------------------------

    def to_ymd(self):
        """ returns the :class:`tuple` of :class:`int` items `(year, month, day)` """
        return self.year, self.month, self.day

    def to_date(self):
        """ returns `datetime.date(year, month, day)` """
        return date(*self.to_ymd())

    def to_float(self):
        """ returns :class:`float` counting the days since Jan, 1st 1900 """
        return float(from_ymd_to_excel(*self.to_ymd()))

    def to_serializable(self, *args, **kwargs):
        return str(self)

    # --- calculation methods ------------------------------------------------

    def _add_days(self, days_int):
        # return super(self.__class__, self).__add__(timedelta(days_int))
        res = super(BaseDateDatetimeDate, self).__add__(timedelta(days_int))
        return self.__class__.from_ymd(res.year, res.month, res.day)

    def _diff_in_days(self, end):
        delta = super(BaseDateDatetimeDate, end).__sub__(self)
        return float(delta.days)



# -*- coding: utf-8 -*-

# businessdate
# ------------
# Python library for generating business dates for fast date operations
# and rich functionality.
# 
# Author:   sonntagsgesicht, based on a fork of Deutsche Postbank [pbrisk]
# Version:  0.5, copyright Wednesday, 18 September 2019
# Website:  https://github.com/sonntagsgesicht/businessdate
# License:  Apache License 2.0 (see LICENSE file)


from datetime import date
from .ymd import is_leap_year


def diff_in_days(start, end):
    """ calculates days between start and end date """
    return float((end-start).days)


def get_30_360(start, end):
    """ implements 30/360 Day Count Convention. """
    start_day = min(start.day, 30)
    end_day = 30 if (start_day == 30 and end.day == 31) else end.day
    return (360 * (end.year - start.year) + 30 * (end.month - start.month) + (end_day - start_day)) / 360.0


def get_30e_360(start, end):
    """ implements the 30E/360 Day Count Convention. """

    y1, m1, d1 = start.timetuple()[:3]
    # adjust to date immediately following the the last day
    y2, m2, d2 = end.timetuple()[:3]

    d1 = min(d1, 30)
    d2 = min(d2, 30)

    return (360 * (y2 - y1) + 30 * (m2 - m1) + (d2 - d1)) / 360.0


def get_30e_360i(start, end):
    """ implements the 30E/360 I. Day Count Convention. """
    y1, m1, d1 = start.timetuple()[:3]
    # adjust to date immediately following the last day
    y2, m2, d2 = end.timetuple()[:3]

    if (m1 == 2 and d1 >= 28) or d1 == 31:
        d1 = 30
    if (m2 == 2 and d2 >= 28) or d2 == 31:
        d2 = 30

    return (360 * (y2 - y1) + 30 * (m2 - m1) + (d2 - d1)) / 360.0


def get_act_360(start, end):
    """ implements Act/360 day count convention. """
    return diff_in_days(start, end) / 360.0


def get_act_365(start, end):
    """ implements Act/365 day count convention. """
    return diff_in_days(start, end) / 365.0


def get_act_36525(start, end):
    """ implements Act/365.25 Day Count Convention """
    return diff_in_days(start, end) / 365.25


def get_act_act(start, end):
    """ implements Act/Act day count convention. """

    # if the period does not lie within a year split the days in the period as following:
    #           remaining days of start year / years in between / days in the end year
    # REMARK: following the before mentioned Definition the first day of the period is included whereas the
    # last day will be excluded
    # What remains to check now is only whether the start and end year are leap or non-leap years. The quotients
    # can be easily calculated and for the years in between they are always one (365/365 = 1; 366/366 = 1)

    if end.year - start.year == 0:
        if is_leap_year(start.year):
            return diff_in_days(start, end) / 366.0  # leap year: 366 days
        return diff_in_days(start, end) / 365.0  # non-leap year: 365 days

    rest_year1 = diff_in_days(start, date(start.year, 12, 31)) + 1  # since the first day counts
    rest_year2 = abs(diff_in_days(end, date(end.year, 1, 1)))  # here the last day is automatically not counted
    years_in_between = end.year - start.year - 1

    return years_in_between + rest_year1 / (366.0 if is_leap_year(start.year) else 365.0) + rest_year2 / (
        366.0 if is_leap_year(end.year) else 365.0)



# -*- coding: utf-8 -*-

# businessdate
# ------------
# Python library for generating business dates for fast date operations
# and rich functionality.
# 
# Author:   sonntagsgesicht, based on a fork of Deutsche Postbank [pbrisk]
# Version:  0.5, copyright Wednesday, 18 September 2019
# Website:  https://github.com/sonntagsgesicht/businessdate
# License:  Apache License 2.0 (see LICENSE file)


from datetime import date, timedelta

from .ymd import easter


class BusinessHolidays(list):
    """ holiday calendar class

    A :class:`BusinessHolidays` instance imitated a list of :class:`datetime.date`
    which can be used to check if a :class:`BusinessDate` is
    included as holiday.

    For convenience input need not to be of type :class:`datetime.date`.
    Duck typing is enough, i.e. having properties
    `year`, `month` and `day`.
    """

    def __init__(self, iterable=()):
        if iterable:
            # iterable = map(BusinessDate, iterable)
            iterable = [bd if isinstance(bd, date) else date(bd.year, bd.month, bd.day) for bd in iterable]
        super(BusinessHolidays, self).__init__(iterable)

    def __contains__(self, item):
        if super(BusinessHolidays, self).__contains__(item):
            return True
        return super(BusinessHolidays, self).__contains__(date(item.year, item.month, item.day))


class TargetHolidays(BusinessHolidays):
    """ holiday calendar class of ecb target2 holidays

    Target holidays are

    * Jan, 1st
    * Good Friday
    * Easter Monday
    * May, 1st
    * December, 25th (Christmas Day)
    * December, 25th (Boxing Day)

    """

    def __contains__(self, item):
        if not super(TargetHolidays, self).__contains__(date(item.year, 1, 1)):
            # add tar days if not done jet

            e = date(*easter(item.year))
            target_days = dict()
            target_days[date(item.year, 1, 1)] = "New Year's Day"
            target_days[e - timedelta(2)] = "Black Friday"
            target_days[e + timedelta(1)] = "Easter Monday"
            target_days[date(item.year, 5, 1)] = "Labour Day"
            target_days[date(item.year, 12, 25)] = "First Christmas Day"
            target_days[date(item.year, 12, 26)] = "Second Christmas Day"

            self.extend(list(target_days.keys()))
        return super(TargetHolidays, self).__contains__(item)



# -*- coding: utf-8 -*-

# businessdate
# ------------
# Python library for generating business dates for fast date operations
# and rich functionality.
# 
# Author:   sonntagsgesicht, based on a fork of Deutsche Postbank [pbrisk]
# Version:  0.5, copyright Wednesday, 18 September 2019
# Website:  https://github.com/sonntagsgesicht/businessdate
# License:  Apache License 2.0 (see LICENSE file)


from .businessperiod import BusinessPeriod
from .businessdate import BusinessDate


class BusinessRange(list):
    def __init__(self, start, stop=None, step=None, rolling=None):
        """ class to build list of business days

        :param BusinessDate start: date to begin schedule,
         if stop not given, start will be used as stop and
         default in rolling to :class:`BusinessDate() <BusinessDate>`
        :param BusinessDate stop: date to stop before,
         if not given, start will be used for stop instead
        :param BusinessPeriod step: period to step schedule,
         if not given 1 day is default
        :param BusinessDate rolling: date to roll on
         (forward and backward) between start and stop,
         if not given default will be start

        **Ansatz** First, `rolling` and `step`
        defines a infinite grid of dates.
        Second, this grid is sliced by `start` (included ,
        if meeting the grid) and `end` (excluded).

        """

        # set default args and build range grid
        start, stop, step, rolling = self._default_args(start, stop, step, rolling)
        schedule = self._build_grid(start, stop, step, rolling)

        # push to super and sort
        super(BusinessRange, self).__init__(set(schedule))
        self.sort()

    @staticmethod
    def _default_args(start, stop, step, rolling):
        if stop is None:
            stop = start
            start = BusinessDate()
        if step is None:
            step = BusinessPeriod(days=1)
        if rolling is None:
            rolling = start
        # make proper businessdate objects
        start = BusinessDate(start)
        rolling = BusinessDate(rolling)
        stop = BusinessDate(stop)
        step = BusinessPeriod(step)
        return BusinessDate(start), BusinessDate(stop), BusinessPeriod(step), BusinessDate(rolling)

    @staticmethod
    def _build_grid(start, stop, step, rolling):
        # setup grid and turn step into positive direction
        grid = list()
        step = step if rolling <= rolling + step else -1 * step

        # roll backward before start
        i = 0
        while start <= rolling + step * i:
            i -= 1

        # fill grid from start until end
        current = rolling + step * i
        while current < stop:
            if start <= current < stop:
                grid.append(current)
            i += 1
            current = rolling + step * i

        return grid

    def adjust(self, convention='', holidays=None):
        """ returns adjusted :class:`BusinessRange` following given convention

        For details of adjusting :class:`BusinessDate` see :meth:`BusinessDate.adjust`.

        For possible conventions invoke :meth:`BusinessDate().adjust() <BusinessDate.adjust>`

        For more details on the conventions see module :mod:`conventions <businessdate.conventions>`)
        """

        adj_list = [d.adjust(convention, holidays) for d in self]
        del self[:]
        super(BusinessRange, self).extend(adj_list)
        return self



# -*- coding: utf-8 -*-

# businessdate
# ------------
# Python library for generating business dates for fast date operations
# and rich functionality.
# 
# Author:   sonntagsgesicht, based on a fork of Deutsche Postbank [pbrisk]
# Version:  0.5, copyright Wednesday, 18 September 2019
# Website:  https://github.com/sonntagsgesicht/businessdate
# License:  Apache License 2.0 (see LICENSE file)


from __future__ import print_function


if False:
    from businessdate import BusinessDate as BD, BusinessSchedule
    from businessdate import BusinessPeriod as BP

    a = BD()
    p = BP('1y2m3d')
    b = a + p

    print(a, p, b - a, a - b)

    a = BD(20160229)
    p = BP('1y1m')
    b = a + p

    print(a, p, b, b - a, a - b)

    print(BD(20160229) + '1y' + '1m', BD(20160229) + '1m' + '1y', BD(20160229) + '1y1m')

    print(not BP("3M") == BP("1M"))
    print(not BP("31d") == BP("1M"))
    print(BP("1Y") == BP("12M"))
    print(BP("3M") == BP("1M") + "2m")

    print(BD("20190418").is_business_day())

if False:
    import businessdate.basedate

    businessdate.basedate.BaseDate = businessdate.basedate.BaseDateFloat

    from businessdate import BusinessDate

    print(businessdate.basedate.BaseDate)
    print('')
    print(BusinessDate().baseclass)
    print(businessdate.BusinessDate().baseclass)
    print(businessdate.businessdate.BusinessDate().baseclass)

if False:
    from time import time
    from businessdate import BusinessDate, BusinessPeriod, BusinessSchedule

    n = 1
    s = BusinessDate()
    e = s + BusinessPeriod('100Y')
    p = BusinessPeriod('1M')

    t = time()
    for _ in range(n):
        BusinessSchedule(s, e, p).adjust()
    print(s.baseclass, 'BusinessSchedule', (time() - t) / n)

    # 1x
    # BaseDateDatetimeDate 0.631363153458
    # BaseDateFloat 0.865006923676
    # BaseDateTuple 1.58643198013

    # 10x
    # BaseDateDatetimeDate 0.640032410622
    # BaseDateFloat 0.852301883698
    # BaseDateTuple 1.23395960331

if False:
    from time import time
    from businessdate import BusinessDate, BusinessPeriod, BusinessSchedule

    n = int(1e6)
    s = BusinessDate()
    e = s + BusinessPeriod('1y')

    t = time()
    for _ in range(n):
        s._add_years(1)
    print(s.baseclass, '_add_years', (time() - t) / n)

    # 1e6 _add_years
    # BaseDateDatetimeDate 4.1959002018e-06
    # BaseDateFloat 6.634775877e-06
    # BaseDateTuple 1.15310940742e-05

    t = time()
    for _ in range(n):
        s._add_months(1)
    print(s.baseclass, '_add_months', (time() - t) / n)

    # 1e6 add_days
    # BaseDateFloat 1.97633004189e-06
    # BaseDateDatetimeDate 5.02036499977e-06
    # BaseDateTuple 1.36047940254e-05

    t = time()
    for _ in range(n):
        s._add_days(1)
    print(s.baseclass, 'add_days', (time() - t) / n)

    # 1e6 add_days
    # BaseDateFloat 1.97633004189e-06
    # BaseDateDatetimeDate 5.02036499977e-06
    # BaseDateTuple 1.36047940254e-05

    t = time()
    for _ in range(n):
        s < e
    print(s.baseclass, 'leq', (time() - t) / n)

    # 1e6 leq
    # BaseDateDatetimeDate 1.44394159317e-07
    # BaseDateFloat 1.7797088623e-07
    # BaseDateTuple 2.16879796982e-06

    t = time()
    for _ in range(n):
        s == e
    print(s.baseclass, '==', (time() - t) / n)

    # 1e6 ==
    # BaseDateDatetimeDate 1.68392896652e-07
    # BaseDateFloat 1.74712896347e-07
    # BaseDateTuple 2.41393589973e-06

    """ 20190724
    BaseDateFloat BusinessSchedule 1.13527107239
    BaseDateFloat _add_years 8.06885409355e-06
    BaseDateFloat _add_months 1.45274348259e-05
    BaseDateFloat add_days 2.22360181808e-06
    BaseDateFloat leq 1.03132009506e-07
    BaseDateFloat == 1.10569000244e-07

    BaseDateDatetimeDate BusinessSchedule 0.504576921463
    BaseDateDatetimeDate _add_years 2.55046105385e-06
    BaseDateDatetimeDate _add_months 3.1582660675e-06
    BaseDateDatetimeDate add_days 5.83482694626e-06
    BaseDateDatetimeDate leq 1.04933023453e-07
    BaseDateDatetimeDate == 1.02378129959e-07
    """

if False:
    from businessdate.businessdate import NewBusinessDate

    n = NewBusinessDate('3M') + NewBusinessDate(20191231)
    print(n)
    print(n.month)
    print(n.months)
    print(n.to_businessdate())

if False:
    from datetime import date, timedelta

    print(date(2015, 0o1, 29) + timedelta(30))
    print(date(2015, 0o1, 29) + timedelta(30) - date(2015, 0o1, 29))

if False:
    from businessdate import BusinessDate as bd

    x = bd('-1M', '20110201', 'MOD_PREVIOUS', bd.DEFAULT_HOLIDAYS)
    print(x, type(x))
    x = bd('3B1Y', None, 'MOD_PREVIOUS', bd.DEFAULT_HOLIDAYS)
    print(x, type(x))
    x = bd('3B1Y1B', None, 'MOD_PREVIOUS', bd.DEFAULT_HOLIDAYS)
    print(x, type(x))
    x = bd('3B-1Y6M-1B', None, 'MOD_PREVIOUS', bd.DEFAULT_HOLIDAYS)
    print(x, type(x))

if False:
    from datetime import date
    from businessdate import BusinessDate

    d = date.fromordinal(735963)
    x = BusinessDate.fromordinal(d.toordinal())
    print(d, x, x.adjust_mod_follow())
    print(BusinessDate.today(), BusinessDate(), date.today())
    print(BusinessDate('0B3D0BMODFOLLOW20171231'))

if False:
    from businessdate import BusinessPeriod

    print(BusinessPeriod(''))
    # print(BusinessPeriod('1B2D1B'))
    print(BusinessPeriod._parse_ymd('1B'))
    print(BusinessPeriod._parse_ymd('1B2D'))
    print(BusinessPeriod._parse_ymd('1B2D1B'))
    print(BusinessPeriod._parse_ymd('2D'))
    print(BusinessPeriod._parse_ymd('2D1B'))

if False:
    from businessdate import BusinessDate

    BusinessDate().adjust()
    BusinessDate().get_day_count(BusinessDate())

    # doctest

    from datetime import date, timedelta
    from businessdate import BusinessDate, BusinessPeriod

    BusinessPeriod(businessdays=10)
    BusinessPeriod(years=2, months=6, days=1)
    BusinessPeriod(years=1, months=6)
    BusinessPeriod(months=18)
    BusinessPeriod(months=1, days=45)
    BusinessPeriod(months=2, days=14)
    BusinessPeriod(months=2, days=15)
    # BusinessPeriod(businessdays=1, days=1)

    from datetime import date, timedelta
    from businessdate import BusinessDate, BusinessPeriod

    a, b = date(2010, 6, 1), date(2010, 12, 31)
    b - a
    BusinessPeriod(b - a)
    timedelta(213)
    BusinessPeriod(timedelta(213))

    from datetime import date, timedelta
    from businessdate import BusinessDate, BusinessPeriod

    BusinessPeriod('-0b')
    BusinessPeriod('-10D')
    BusinessPeriod('-1y3m4d')
    BusinessPeriod('-18M')
    BusinessPeriod('-1Q')
    BusinessPeriod('-2w')
    BusinessPeriod('-10B')
    BusinessPeriod(years=-2, months=-6, days=-1)

    from businessdate import BusinessDate, BusinessPeriod

    BusinessDate.BASE_DATE = 20161010
    BusinessDate()
    # create from period
    BusinessDate(BusinessPeriod(months=1))
    # create from period string
    BusinessDate('1m')
    BusinessDate() + '1m'
    # works with additional date as well
    BusinessDate('1m20161213')
    BusinessDate('20161213') + '1m'
    # even for business days
    BusinessDate('15b')
    BusinessDate() + '15b'
    # add adjustment convention 'end_of_month' with a business date
    BusinessDate('0bEOM')
    BusinessDate('15bEOM')
    # add adjustment convention 'mod_follow' with a business date
    BusinessDate('0bModFlw')
    BusinessDate('15bModFlw')
    # adjustment convention without a business date is ignored
    BusinessDate('EOM')
    BusinessDate('ModFlw')
    # even together with a classical period since the adjustment statement is ambiguous
    # should the start date (spot) or end date be adjusted?
    BusinessDate('1mModFlw')
    # but adding zero business days clarifies it
    BusinessDate('0b1mModFlw')
    BusinessDate('0b1mModFlw') == BusinessDate().adjust('ModFlw') + '1m'
    BusinessDate('1m0bModFlw')
    BusinessDate('1m0bModFlw') == (BusinessDate() + '1m').adjust('ModFlw')
    # clearly business days may be non zero, too
    BusinessDate('15b1mModFlw')
    BusinessDate('15b1mModFlw') == BusinessDate('15b').adjust('ModFlw') + '1m'
    BusinessDate('1m5bModFlw')
    BusinessDate('1m5bModFlw') == (BusinessDate() + '1m').adjust('ModFlw') + '5b'
    # putting all together we get
    BusinessDate('15b1m5bModFlw20161213')
    BusinessDate('15b1m5bModFlw20161213') == ((BusinessDate(20161213) + '15b').adjust('ModFlw') + '1m').adjust(
        'ModFlw') + '5b'

if False:
    from businessdate import BusinessDate, BusinessPeriod, BusinessRange

    rng = BusinessRange('19990101', '20051231')
    for a in rng:
        for b in rng:
            if min(a, b).day > 0:
                b - a

if False:
    from datetime import timedelta, date
    from businessdate import BusinessDate, BusinessPeriod, BusinessRange

    for p in ('1M0D', '1M1D', '1M2D', '1M3D', '100D'):
        jan = BusinessDate(20150129)
        mar = jan + p
        print(('mar = jan + p', p, jan, mar, mar-jan, jan-mar))

    for p in ('1M0D', '1M1D', '1M2D', '1M3D', '100D'):
        mar = BusinessDate(20150301)
        jan = mar - p
        print(('jan = mar - p', p, jan, mar, mar-jan, jan-mar))

if False:

    class avg_day_per_month(object):

        cum_days = list(enumerate([31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]))

        def __init__(self, max=12, start=0):
            self.start = start
            self.pos = self.start
            self.cum = 0
            self.num = 0
            self.max = max


        def __iter__(self):
            self.pos = self.start
            self.cum = 0
            self.num = 0
            return self

        def __next__(self):
            if self.max <= self.num:
                raise StopIteration
            i, d = self.__class__.cum_days[self.pos % 12]
            self.pos += 1
            self.cum += (d + 1) if self.pos % 48 == 36 + 2 else d
            self.num += 1
            return self.cum

    # for i in range(124):
    #     starts = [list(iter(avg_day_per_month(max=i+1, start=s)))[-1] for s in range(12)]
    #     mx = [i+1 for i, e in enumerate(starts) if e == max(starts)]
    #     mn = [i+1 for i, e in enumerate(starts) if e == min(starts)]
    #     # print((i, max(starts), mx))
    #     # print((i, min(starts), mn))
    #     #print((i+1, min(starts), max(starts), max(starts)-min(starts)))
    #     #print((i, max(starts), min(starts)))
    #     #print(min(starts))

    # days = 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31, 31
    # rday = tuple(reversed(days))
    # print(rday)
    # mn,mx = 0, 0
    # for i in range(12):
    #     mn += days[int(i%12)]
    #     mn += 1 if int(i % 48) == 36 else 0
    #     mx += rday[int(i%12)]
    #     mx += 1 if int(i % 48) == 11 else 0
    #     print((i+1, mn, mx, mx-mn))

    from businessdate import BusinessPeriod, BusinessDate, BusinessRange

    jan31 = BusinessDate(20010131)
    feb01 = BusinessDate(20010201)

    for date in BusinessRange(20000101, 21000101):
        if date.month==1 and date.day==1:
            print('\n:', end='')
        for m in range(12):
            period = BusinessPeriod(months=m)
            mn, mx = period.min_days(), period.max_days()
            # dn, dx = feb01.diff_in_days(feb01 + period), (jan31 - period).diff_in_days(jan31)
            fwd, bck = date.diff_in_days(date + period), (date - period).diff_in_days(date)
            if not mn < fwd < mx:
                print('.', end='')
                if not mn <= fwd <= mx:
                    print((period, date, mn, fwd, mx))
            elif not mn < bck < mx:
                print('*', end='')
                if not mn <= bck <= mx:
                    print((period, date, mn, bck, mx))
            print(' ', end='')

if False:
    from businessdate import BusinessPeriod, BusinessDate, BusinessRange

    BusinessPeriod('13m').max_days()  # 397
    BusinessPeriod('13m').min_days()  # 393
    BusinessPeriod('13m') < BusinessPeriod('395d')  # not well defined -> None
    BusinessPeriod('13m') < BusinessPeriod('397d') # False
    BusinessPeriod('13m') <= BusinessPeriod('397d') # True
    BusinessPeriod('ON') == BusinessPeriod('1B') # True
    BusinessPeriod('7D') == BusinessPeriod('1W') # True
    BusinessPeriod('30D') == BusinessPeriod('1M') # False
    BusinessPeriod('1D') == BusinessPeriod('1B')  # not well defined -> None

if True:

    from businessdate import BusinessDate
    s, e = BusinessDate(20190915), BusinessDate(20190915) + '10y'
    for a in dir(s):
        print(a)

    print(s.get_act_360(e))
    print(s.get_day_count(e, 'act360'))
    print(s.get_act_act(e))
    print(s.get_day_count(e, 'actact'))
    print(s.adjust_flw())
    print(s.adjust_follow())
    print(s.adjust('follow'))
    print(s.adjust_imm())
    print(s.adjust('imm'))



# -*- coding: utf-8 -*-

# businessdate
# ------------
# Python library for generating business dates for fast date operations
# and rich functionality.
# 
# Author:   sonntagsgesicht, based on a fork of Deutsche Postbank [pbrisk]
# Version:  0.5, copyright Wednesday, 18 September 2019
# Website:  https://github.com/sonntagsgesicht/businessdate
# License:  Apache License 2.0 (see LICENSE file)


from calendar import WEDNESDAY, FRIDAY
from datetime import date, timedelta
from .ymd import days_in_month, end_of_quarter_month

ONE_DAY = timedelta(1)


def is_business_day(business_date, holidays=list()):
    """ method to check if a date falls neither on weekend nor is in holidays. """

    if business_date.weekday() > FRIDAY:
        return False
    return business_date not in holidays


def adjust_no(business_date, holidays=()):
    """ does no adjustment. """
    return business_date


def adjust_previous(business_date, holidays=()):
    """ adjusts to Business Day Convention "Preceding". """
    while not is_business_day(business_date, holidays):
        business_date -= ONE_DAY
    return business_date


def adjust_follow(business_date, holidays=()):
    """ adjusts to Business Day Convention "Following". """
    while not is_business_day(business_date, holidays):
        business_date += ONE_DAY
    return business_date


def adjust_mod_follow(business_date, holidays=()):
    """ adjusts to Business Day Convention "Modified [Following]". """
    month = business_date.month
    new = adjust_follow(business_date, holidays)
    if month != new.month:
        new = adjust_previous(business_date, holidays)
    business_date = new
    return business_date


def adjust_mod_previous(business_date, holidays=()):
    """ adjusts to Business Day Convention "Modified Preceding". """
    month = business_date.month
    new = adjust_previous(business_date, holidays)
    if month != new.month:
        new = adjust_follow(business_date, holidays)
    business_date = new
    return business_date


def adjust_start_of_month(business_date, holidays=()):
    """ adjusts to Business Day Convention "Start of month", i.e. first business day. """
    business_date = date(business_date.year, business_date.month, 1)
    business_date = adjust_follow(business_date, holidays)
    return business_date


def adjust_end_of_month(business_date, holidays=()):
    """ adjusts to Business Day Convention "End of month", i.e. last business day. """
    y, m, d = business_date.year, business_date.month, business_date.day
    business_date = date(y, m, days_in_month(y, m))
    business_date = adjust_previous(business_date, holidays)
    return business_date


def adjust_imm(business_date, holidays=()):
    """ adjusts to Business Day Convention of "International Monetary Market". """
    business_date = date(business_date.year, end_of_quarter_month(business_date.month), 15)
    while business_date.weekday() == WEDNESDAY:
        business_date += ONE_DAY
    return business_date


def adjust_cds_imm(business_date, holidays=()):
    """ adjusts to Business Day Convention "Single Name CDS". """
    business_date = date(business_date.year, end_of_quarter_month(business_date.month), 20)
    return business_date



# -*- coding: utf-8 -*-

# businessdate
# ------------
# Python library for generating business dates for fast date operations
# and rich functionality.
# 
# Author:   sonntagsgesicht, based on a fork of Deutsche Postbank [pbrisk]
# Version:  0.5, copyright Wednesday, 18 September 2019
# Website:  https://github.com/sonntagsgesicht/businessdate
# License:  Apache License 2.0 (see LICENSE file)


from datetime import timedelta


class BusinessPeriod(object):

    def __init__(self, period='', years=0, quarters=0, months=0, weeks=0, days=0, businessdays=0):
        """ class to store and calculate date periods as combinations of days, weeks, years etc.

        :param str period: encoding a business period.
         Such is given by a sequence of digits as :class:`int` followed by a :class:`char` -
         indicating the number of
         years **Y**,
         quarters **Q** (which is equivalent to 3 month),
         month **M**,
         weeks **W** (which is equivalent to 7 days),
         days **D**,
         business days **B**.
         E.g. **1Y2W3D** what gives a period of 1 year plus 2 weeks and 3 days
         (see :doc:`tutorial <tutorial>` for details).

        :param int years: number of years in the period (equivalent to 12 months)
        :param int quarters: number of quarters in the period (equivalent to 3 months)
        :param int months: number of month in the period
        :param int weeks: number of weeks in the period (equivalent to 7 days)
        :param int days: number of days in the period
        :param int businessdays: number of business days,
         i.e. days which are neither weekend nor :class:`holidays <BusinessHolidays>`,  in the period.
         Only either `businessdays` or the others can be given.
         Both at the same time is not allowed.

        """
        if period and any((years, months, days, businessdays)):
            raise ValueError(
                "Either string or argument input only for %s" % self.__class__.__name__)

        super(BusinessPeriod, self).__init__()
        if isinstance(period, BusinessPeriod):
            years = period.years
            months = period.months
            days = period.days
            businessdays = period.businessdays
        elif isinstance(period, timedelta):
            days = period.days
        elif period is None:
            pass
        elif isinstance(period, str):
            if period.upper() == '':
                pass
            elif period.upper() == '0D':
                pass
            elif period.upper() == 'ON':
                businessdays = 1
            elif period.upper() == 'TN':
                businessdays = 2
            elif period.upper() == 'DD':
                businessdays = 3
            else:
                s, y, q, m, w, d, f = BusinessPeriod._parse_ymd(period)
                # no final businesdays allowed
                if f:
                    raise ValueError("Unable to parse %s as %s" % (period, self.__class__.__name__))
                # except the first non vanishing of y,q,m,w,d must have positive sign
                sgn = [int(x / abs(x)) for x in (y, q, m, w, d) if x]
                if [x for x in sgn[1:] if x < 0]:
                    raise ValueError(
                        "Except at the beginning no signs allowed in %s as %s" % (str(period), self.__class__.__name__))
                y, q, m, w, d = (abs(x) for x in (y, q, m, w, d))
                # use sign of first non vanishing of y,q,m,w,d
                sgn = sgn[0] if sgn else 1
                businessdays, years, quarters, months, weeks, days = s, sgn * y, sgn * q, sgn * m, sgn * w, sgn * d
        else:
            raise TypeError(
                "%s of Type %s not valid to create BusinessPeriod." %(str(period), period.__class__.__name__))

        self._months = 12 * years + 3 * quarters + months
        self._days = 7 * weeks + days
        self._businessdays = businessdays

        if businessdays and (self._months or self._days):
            raise ValueError(
                "Either (years,months,days) or businessdays must be zero for %s" % self.__class__.__name__)
        if self._months and not self._days / self._months >= 0:
            ymd = self.years, self.months, self.days
            raise ValueError(
                "(years, months, days)=%s must have equal sign for %s" % (str(ymd), self.__class__.__name__))

    @property
    def years(self):
        return int(-1 * (-1 * self._months // 12) if self._months < 0 else self._months // 12)

    @property
    def months(self):
        return int(-1 * (-1 * self._months % 12) if self._months < 0 else self._months % 12)

    @property
    def days(self):
        return int(self._days)

    @property
    def businessdays(self):
        return int(self._businessdays)

    # --- validation and information methods ---------------------------------

    @classmethod
    def _parse_ymd(cls, period):
        # can even parse strings like '-1B-2Y-4Q+5M' but also '0B', '-1Y2M3D' as well.
        period = period.upper().replace(' ', '')
        period = period.replace('BUSINESSDAYS', 'B')
        period = period.replace('YEARS', 'Y')
        period = period.replace('QUARTERS', 'Q')
        period = period.replace('MONTHS', 'M')
        period = period.replace('WEEKS', 'W')
        period = period.replace('DAYS', 'D')

        def _parse(p, letter):
            if p.find(letter) >= 0:
                s, p = p.split(letter, 1)
                s = s[1:] if s.startswith('+') else s
                sgn, s = (-1, s[1:]) if s.startswith('-') else (1, s)
                if not s.isdigit():
                    raise ValueError("Unable to parse %s in %s as %s" % (s, p, cls.__name__))
                return sgn * int(s), p
            return 0, p

        p = period.upper()

        # p[-1] is not 'B', p.strip('0123456789+-B')==''
        s, p = _parse(p, 'B') if not p[-1]=='B' else (0, p)
        s, p = _parse(p, 'B') if not p.strip('0123456789+-B') else (s, p)
        s, p = _parse(p, 'B') if p.count('B') > 1 else (s, p)
        y, p = _parse(p, 'Y')
        q, p = _parse(p, 'Q')
        m, p = _parse(p, 'M')
        w, p = _parse(p, 'W')
        d, p = _parse(p, 'D')
        f, p = _parse(p, 'B')
        if not p == '':
            raise ValueError("Unable to parse %s as %s" % (p, cls.__name__))
        return s, y, q, m, w, d, f

    @classmethod
    def is_businessperiod(cls, period):
        """ returns true if the argument can be understood as :class:`BusinessPeriod` """
        if period is None:
            return False
        if isinstance(period, (int, float, list, set, dict, tuple)):
            return False
        if isinstance(period, (timedelta, BusinessPeriod)):
            return True
        if period in ('', '0D', 'ON', 'TN', 'DD'):
            return True
        if isinstance(period, str):
            if period.isdigit():
                return False
            #if period.upper().strip('+-0123456789BYQMWD'):
            #    return False
            try:  # to be removed
                BusinessPeriod._parse_ymd(period)
            except ValueError:
                return False
            return True
        return False

    # --- operator methods ---------------------------------------------------

    def __repr__(self):
        return self.__class__.__name__ + "('%s')" % str(self)

    def __str__(self):

        if self.businessdays:
            period_str = str(self.businessdays) + 'B'
        else:
            period_str = '-' if self.years < 0 or self.months < 0 or self.days < 0 else ''
            if self.years:
                period_str += str(abs(self.years)) + 'Y'
            if self.months:
                period_str += str(abs(self.months)) + 'M'
            if self.days:
                period_str += str(abs(self.days)) + 'D'

        if not period_str:
            period_str = '0D'
        return period_str

    def __abs__(self):
        ymdb = self.years, self.months, self.days, self.businessdays
        y,m,d,b = tuple(map(abs, ymdb))
        return self.__class__(years=y, months=m, days=d, businessdays=b)

    def __cmp__(self, other):
        other = self.__class__() if other == 0 else other
        if not isinstance(other, BusinessPeriod):
            other = BusinessPeriod(other)
        if self.businessdays:
            if other and not other.businessdays:
                # log warning on non compatible pair
                return None
            return self.businessdays - other.businessdays
        m = 12 * (self.years - other.years) + self.months - other.months
        d = self.days - other.days
        if m * 28 < -d < m * 31:
            p = self.__class__(months=m)
            if p.min_days() <= -d <= p.max_days():
                # log warning on non orderable pair
                return None
        return m * 30.5 + d

    def __eq__(self, other):
        if isinstance(other, type(self)):
            attr = 'years', 'months', 'days', 'businessdays'
            return all(getattr(self, a) == getattr(other, a) for a in attr)
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __le__(self, other):
        cmp = self.__cmp__(other)
        cmp = self.__cmp__(other + '1D') if cmp is None else cmp
        return cmp if cmp is None else cmp <= 0

    def __lt__(self, other):
        cmp = self.__cmp__(other)
        return cmp if cmp is None else cmp < 0

    def __ge__(self, other):
        lt = self.__lt__(other)
        return None if lt is None else not lt

    def __gt__(self, other):
        le = self.__le__(other)
        return None if le is None else not le

    def __hash__(self):
        return hash(repr(self))

    def __nonzero__(self):
        # return any((self.years, self.months, self.days, self.businessdays))
        return self.__bool__()

    def __bool__(self):
        # return self.__nonzero__()
        return any((self._months, self._days, self._businessdays))

    def __add__(self, other):
        if isinstance(other, (list, tuple)):
            return [self + o for o in other]
        if BusinessPeriod.is_businessperiod(other):
            p = BusinessPeriod(other)
            y = self.years + p.years
            m = self.months + p.months
            d = self.days + p.days
            b = self.businessdays + p.businessdays
            return self.__class__(years=y, months=m, days=d, businessdays=b)
        raise TypeError('addition of BusinessPeriod cannot handle objects of type %s.' % other.__class__.__name__)

    def __sub__(self, other):
        if isinstance(other, (list, tuple)):
            return [self - o for o in other]
        if BusinessPeriod.is_businessperiod(other):
            return self + (-1 * BusinessPeriod(other))
        raise TypeError('subtraction of BusinessPeriod cannot handle objects of type %s.' % other.__class__.__name__)

    def __mul__(self, other):
        if isinstance(other, (list, tuple)):
            return [self * o for o in other]
        if isinstance(other, int):
            m = other * self._months
            d = other * self._days
            b = other * self._businessdays
            return BusinessPeriod(months=m, days=d, businessdays=b)
        raise TypeError("expected int type but got %s" % other.__class__.__name__)

    def __rmul__(self, other):
        return self.__mul__(other)

    def max_days(self):
        if self._months < 0:
            sgn = -1
            days_in_month = 31, 30, 31, 30, 31, 31, 30, 31, 30, 31, 31, 28  # days from mar to feb forwards
        else:
            sgn = 1
            days_in_month = 31, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31, 28
        m = sgn * self._months
        # days from jan to feb backwards
        days = 0
        for i in range(m):
            days += days_in_month[int(i % 12)]
            days += 1 if int(i % 48) == 11 else 0
        return sgn * days + self._days

    def min_days(self):
        if self._months < 0 :
            sgn = -1
            days_in_month = 28, 31, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 # days from feb to jan backwards
        else:
            sgn = 1
            days_in_month = 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31, 31 # days from feb to jan forwards
        m  = sgn * self._months
        days = 0
        for i in range(m):
            days += days_in_month[int(i % 12)]
            days += 1 if int(i % 48) == 36 else 0
        return sgn * days + self._days



# -*- coding: utf-8 -*-

# businessdate
# ------------
# Python library for generating business dates for fast date operations
# and rich functionality.
# 
# Author:   sonntagsgesicht, based on a fork of Deutsche Postbank [pbrisk]
# Version:  0.5, copyright Wednesday, 18 September 2019
# Website:  https://github.com/sonntagsgesicht/businessdate
# License:  Apache License 2.0 (see LICENSE file)


from datetime import date, datetime, timedelta

from . import conventions
from . import daycount
from .ymd import is_leap_year, days_in_year, days_in_month, end_of_quarter_month
from .basedate import BaseDateFloat, BaseDateDatetimeDate
from .businessholidays import TargetHolidays
from .businessperiod import BusinessPeriod


class BusinessDate(BaseDateDatetimeDate):
    ADJUST = 'No'
    BASE_DATE = None
    DATE_FORMAT = '%Y%m%d'
    DAY_COUNT = 'act_36525'
    DEFAULT_HOLIDAYS = TargetHolidays()

    _adj_func = {
        'no': conventions.adjust_no,
        'previous': conventions.adjust_previous,
        'prev': conventions.adjust_previous,
        'prv': conventions.adjust_previous,
        'mod_previous': conventions.adjust_mod_previous,
        'modprevious': conventions.adjust_mod_previous,
        'modprev': conventions.adjust_mod_previous,
        'modprv': conventions.adjust_mod_previous,
        'follow': conventions.adjust_follow,
        'flw': conventions.adjust_follow,
        'modified': conventions.adjust_mod_follow,
        'mod_follow': conventions.adjust_mod_follow,
        'modfollow': conventions.adjust_mod_follow,
        'modflw': conventions.adjust_mod_follow,
        'start_of_month': conventions.adjust_start_of_month,
        'startofmonth': conventions.adjust_start_of_month,
        'som': conventions.adjust_start_of_month,
        'end_of_month': conventions.adjust_end_of_month,
        'endofmonth': conventions.adjust_end_of_month,
        'eom': conventions.adjust_end_of_month,
        'imm': conventions.adjust_imm,
        'cds_imm': conventions.adjust_cds_imm,
        'cdsimm': conventions.adjust_cds_imm,
        'cds': conventions.adjust_cds_imm,
    }
    _dc_func = {
        '30_360': daycount.get_30_360,
        '30360': daycount.get_30_360,
        'thirty360': daycount.get_30_360,
        '30e_360': daycount.get_30e_360,
        '30e360': daycount.get_30e_360,
        'thirtye360': daycount.get_30e_360,
        '30e_360_i': daycount.get_30e_360i,
        '30e360i': daycount.get_30e_360i,
        'thirtye360i': daycount.get_30e_360i,
        'act_360': daycount.get_act_360,
        'act360': daycount.get_act_360,
        'act_365': daycount.get_act_365,
        'act365': daycount.get_act_365,
        'act_36525': daycount.get_act_36525,
        'act_365.25': daycount.get_act_36525,
        'act36525': daycount.get_act_36525,
        'act_act': daycount.get_act_act,
        'actact': daycount.get_act_act,
    }

    def __new__(cls, year=None, month=0, day=0, convention=None, holidays=None):
        """ date class to perform calculations coming from financial businesses

        :param year: number of year or some other input value t
         o create :class:`BusinessDate` instance.
         When applying other input, this can be either
         :class:`int`, :class:`float`, :class:`datetime.date` or :class:`string`
         which will be parsed and transformed into equivalent
         :class:`tuple` of :class:`int` items `(year,month,day)`
         (See :doc:`tutorial <tutorial>` for details).
        :param int month: number of month in year 1 ... 12
         (default: 0, required to be 0 when other input of year is used)
        :param int days: number of day in month 1 ... 31
         (default: 0, required to be 0 when other input of year is used)

        For all input arguments exits read only properties.

        """

        '''
        :param str convention: keyword to select a business day adjustment convention
         which is used as default for :meth:`BusinessDate.adjust`.
         For more details on the conventions see module :mod:`businessdate.conventions`.
        :param list holidays: container containing items of type :class:`datetime.date`
         which is used as default for :meth:`BusinessDate.adjust`.
        '''

        if isinstance(year, str):
            year, month, day = cls._parse_date_string(year, default=(year, month, day))

        if isinstance(year, (date, BaseDateFloat, BaseDateDatetimeDate, BusinessDate)):
            year, month, day = year.year, year.month, year.day

        if isinstance(year, (int, float)) and 10000101 <= year:  # start 20191231 representation from 1000 a.d.
            ymd = str(year)
            year, month, day = int(ymd[:4]), int(ymd[4:6]), int(ymd[6:])

        if isinstance(year, int) and month and day:
            if 12 < month:
                year += int(month // 12)
                month = int(month % 12)
            if issubclass(cls, BaseDateFloat):
                return cls.from_ymd(year, month, day)
            return super(BusinessDate, cls).__new__(cls, year, month, day)

        if isinstance(year, (int, float)) and 1 < year < 10000101:  # excel representation before 1000 a.d.
            if issubclass(cls, BaseDateDatetimeDate):
                return cls.from_float(year)
            return super(BusinessDate, cls).__new__(cls, year)

        if isinstance(year, (list, tuple)):
            return list(map(BusinessDate, year))

        if year is None:
            if cls.BASE_DATE is None:
                return cls(date.today())
            return cls(cls.BASE_DATE)

        if isinstance(year, timedelta):
            year = '%sD' % year.days

        # try to split complex or period input, e.g. '0B1D2BMOD20191231' or '3Y2M1D' or '-2B'
        return cls._from_complex_input(str(year))

    @classmethod
    def _parse_date_string(cls, date_str, default=None):
        date_str = str(date_str)
        if date_str.count('-'):
            str_format = '%Y-%m-%d'
        elif date_str.count('.'):
            str_format = '%d.%m.%Y'
        elif date_str.count('/'):
            str_format = '%m/%d/%Y'
        elif len(date_str) == 8 and date_str.isdigit():
            str_format = '%Y%m%d'
        else:
            str_format = ''
        if str_format:
            date_date = datetime.strptime(date_str, str_format)
            return date_date.year, date_date.month, date_date.day

        if default is None:
            raise ValueError("The input %s has not the right format for %s" % (date_str, cls.__name__))
        return default

    @classmethod
    def _from_complex_input(cls, date_str):
        date_str = str(date_str).upper()
        convention, origin, holidays = None, None, None

        # first, extract origin
        if len(date_str) > 8:
            try:
                datetime.strptime(date_str[-8:], '%Y%m%d')
                origin = date_str[-8:]
                date_str = date_str[:-8]
            except ValueError:
                # no date found a the end of the string
                pass

        # second, extract convention
        for a in sorted(cls._adj_func.keys(), key=len, reverse=True):
            if date_str.find(a.upper()) >= 0:
                convention = a
                date_str = date_str[:-len(a)]
                break
        if not date_str:
            date_str = '0B'

        # third, parse spot, period and final
        pfields = date_str.strip('0123456789+-B')
        spot, period, final = date_str, '', ''
        if pfields:
            spot, period, final = '', '', ''
            x = pfields[-1]
            period, final = date_str.split(x, 1)
            period += x
            if period.find('B') >= 0:
                spot, period = period.split('B', 1)
                spot += 'B'

        # third, build BusinessDate and adjust by conventions to periods
        res = cls(origin)
        if spot:
            if convention:
                res = res.adjust(convention, holidays)
            res = res.add_period(spot, holidays)
        if period:
            res = res.add_period(period, holidays)
        if final:
            if convention:
                res = res.adjust(convention, holidays)
            res = res.add_period(final, holidays)
        return res

    @classmethod
    def is_businessdate(cls, d):
        """ checks whether the provided input can be a date """
        if not isinstance(d, (date, BaseDateFloat, BaseDateDatetimeDate)):
            try:  # to be removed
                cls(d)
            except ValueError:
                return False
        return True

    def __copy__(self):
        return self.__deepcopy__()

    def __deepcopy__(self, memodict={}):
        return BusinessDate(date(*self.to_ymd()))

    # --- operator methods ---------------------------------------------------

    def __add__(self, other):
        """
            addition of BusinessDate.

        :param other: can be BusinessPeriod or
        any thing that might be casted to it. Or a list of them.
        """
        if isinstance(other, (list, tuple)):
            return [self + pd for pd in other]
        if BusinessPeriod.is_businessperiod(other):
            return self.add_period(other)
        raise TypeError('addition of BusinessDates cannot handle objects of type %s.' % other.__class__.__name__)

    def __sub__(self, other):
        """
            subtraction of BusinessDate.

        :param other: can be other BusinessDate, BusinessPeriod or
        any thing that might be casted to those. Or a list of them.
        """
        if isinstance(other, (list, tuple)):
            return [self - pd for pd in other]
        if BusinessPeriod.is_businessperiod(other):
            return self + (-1 * BusinessPeriod(other))
        if BusinessDate.is_businessdate(other):
            y, m, d = BusinessDate(other).diff_in_ymd(self)
            return BusinessPeriod(years=y, months=m, days=d)
        raise TypeError('subtraction of BusinessDates cannot handle objects of type %s.' % other.__class__.__name__)

    def __str__(self):
        date_format = self.__class__.DATE_FORMAT
        return self.to_date().strftime(date_format)

    def __repr__(self):
        return self.__class__.__name__ + "(%s)" % str(self)

    # --- validation and information methods ------------------------

    def is_leap_year(self):
        """ returns `True` for leap year and False otherwise """
        return is_leap_year(self.year)

    def days_in_year(self):
        """ returns number of days in the given calendar year """
        return days_in_year(self.year)

    def days_in_month(self):
        """ returns number of days for the month """
        return days_in_month(self.year, self.month)

    def end_of_month(self):
        """ returns the day of the end of the month as :class:`BusinessDate` object"""
        return BusinessDate(self.year, self.month, self.days_in_month())

    def end_of_quarter(self):
        """ returns the day of the end of the quarter as :class:`BusinessDate` object"""
        return BusinessDate(self.year, end_of_quarter_month(self.month), 0o1).end_of_month()

    def is_business_day(self, holidays=None):
        """ returns `True` if date falls neither on weekend nor is in holidays (if given as container object) """
        holidays = self.__class__.DEFAULT_HOLIDAYS if holidays is None else holidays
        return conventions.is_business_day(self.to_date(), holidays)

    # --- calculation methods --------------------------------------------

    def _add_business_days(self, days_int, holidays=None):
        holidays = self.__class__.DEFAULT_HOLIDAYS if holidays is None else holidays

        res = self.__deepcopy__()
        if days_int >= 0:
            count = 0
            while count < days_int:
                res = res._add_days(1)
                if res.is_business_day(holidays):
                    count += 1
        else:
            count = 0
            while count > days_int:
                res = res._add_days(-1)
                if res.is_business_day(holidays):
                    count -= 1

        return res

    def _add_ymd(self, years=0, months=0, days=0):
        s = self
        y = s.year + years
        m = s.month + months
        while m < 1:
            m += 12
            y -= 1
        som = self.__class__(y, m, 1)
        d = min(self.day, som.days_in_month()) - 1 + days
        return som._add_days(d)

    def add_period(self, period_obj, holidays=None):
        """ adds a :class:`BusinessPeriod` object
        or anythings that create one and returns :class:`BusinessDate` object.

        It is simply adding the number of `years`, `months` and `days` or
        if `businessdays` given the number of business days,
        i.e. days neither weekend nor in holidays (see also :meth:`BusinessDate.is_business_day`)
        """

        p = BusinessPeriod(period_obj)
        res = self
        res = res._add_business_days(p.businessdays, holidays)
        res = res._add_ymd(p.years, p.months, p.days)
        return res

    def diff_in_days(self, end_date):
        """ calculates the distance to a :class:`BusinessDate` in days """
        return int(self._diff_in_days(end_date))

    def diff_in_ymd(self, end_date):

        if end_date < self:
            y,m,d = 0,0,0
            while end_date < self._add_ymd(y,0,0):
                y -= 1
            while end_date < self._add_ymd(y+1,m,0):
                m -= 1
            while end_date < self._add_ymd(y+1,m+1,d):
                d -= 1
            return y+1, m+1, d

        y = end_date.year - self.year
        m = end_date.month - self.month
        d = end_date.day - self.day

        while m < 0:
            y -= 1
            m += 12

        while d < 0:
            m -= 1
            if m < 0:
                y -= 1
                m += 12
            d = self._add_ymd(y, m, 0)._diff_in_days(end_date)

        # if not any((0 <= y, 0 <= m < 12, 0 <= d < 31)):
        #     raise AssertionError((y,m,d))
        # if not end_date == self._add_ymd(y, m, d):
        #     raise AssertionError('%s!=%s==%s._add_ymd(%d,%d,%d)' %
        #                    (end_date, self._add_ymd(y, m, d), self, y, m, d))
        # if not self == end_date._add_ymd(-y, -m, -d) and min(end_date, self).day < 29:
        #     raise AssertionError('%s!=%s==%s._add_ymd(%d,%d,%d)' %
        #                    (self, end_date._add_ymd(-y, -m, -d), end_date, -y,-m,-d))

        return int(y), int(m), int(d)

    # --- business day adjustment and day count fraction methods -----------------------------------------

    def get_day_count(self, end=None, convention=''):
        """ counts the days as a year fraction to given date following the specified convention.

        For more details on the conventions see module :mod:`businessdate.daycount`.
        """
        convention = convention if convention else BusinessDate.DAY_COUNT
        dc_func = self.__class__._dc_func
        return dc_func[convention.lower()](self.to_date(), BusinessDate(end).to_date())

    def get_year_fraction(self, end=None, convention=''):
        """ wrapper for :meth:`BusinessDate.get_day_count` method for different naming preferences """
        return self.get_day_count(end, convention)

    def adjust(self, convention='', holidays=None):
        """ returns an adjusted :class:`BusinessDate` if it was not a business day following the specified convention.

        For details on business days see :meth:`BusinessDate.is_business_day`.

        For more details on the conventions see module :mod:`businessdate.conventions`
        """
        convention = convention if convention else BusinessDate.ADJUST
        adj_func = self.__class__._adj_func
        holidays = self.__class__.DEFAULT_HOLIDAYS if holidays is None else holidays
        return BusinessDate(adj_func[convention.lower()](self.to_date(), holidays))


    def __getattr__(self, item):
        if item.startswith('adjust_'):
            return  lambda h=None: self.adjust(item.replace('adjust_',''), h)
        if item.startswith('get_'):
            return  lambda e: self.get_year_fraction(e, item.replace('get_',''))
        raise AttributeError("'%s' object has no attribute '%s'" % (self.__class__.__name__, item))


# add additional __doc__ at runtime (during import)
try:
    s = '\n' \
        '        In order to get the year fraction according a day count convention \n' \
        '        provide one of the following convention key words: \n\n'
    for k, v in BusinessDate._dc_func.items():
         s +='           * ' + (":code:`%s`" % k).ljust(16) + '' + v.__doc__ + '\n\n'
    BusinessDate.get_day_count.__doc__ +=s

    s = '\n' \
        '        In order to adjust according a business day convention \n' \
        '        provide one of the following convention key words: \n\n'
    for k, v in BusinessDate._adj_func.items():
        s += '           * ' + (":code:`%s`" % k).ljust(16) + '' + v.__doc__ + '\n\n'
    BusinessDate.adjust.__doc__ +=s

    del s
except AttributeError:
    pass



# -*- coding: utf-8 -*-

# businessdate
# ------------
# Python library for generating business dates for fast date operations
# and rich functionality.
# 
# Author:   sonntagsgesicht, based on a fork of Deutsche Postbank [pbrisk]
# Version:  0.5, copyright Wednesday, 18 September 2019
# Website:  https://github.com/sonntagsgesicht/businessdate
# License:  Apache License 2.0 (see LICENSE file)


__doc__ = 'Python library for generating business dates for fast date operations and rich functionality.'
__version__ = '0.5'
__dev_status__ = '4 - Beta'
__date__ = 'Wednesday, 18 September 2019'
__author__ = 'sonntagsgesicht, based on a fork of Deutsche Postbank [pbrisk]'
__email__ = 'sonntagsgesicht@icloud.com'
__url__ = 'https://github.com/sonntagsgesicht/' + __name__
__license__ = 'Apache License 2.0'
__dependencies__ = ()
__dependency_links__ = ()
__data__ = ()
__scripts__ = ()

import logging
logging.getLogger(__name__).addHandler(logging.NullHandler())

from .businessholidays import BusinessHolidays
from .businessperiod import BusinessPeriod
from .businessdate import BusinessDate
from .businessrange import BusinessRange
from .businessschedule import BusinessSchedule



# -*- coding: utf-8 -*-

# businessdate
# ------------
# Python library for generating business dates for fast date operations
# and rich functionality.
# 
# Author:   sonntagsgesicht, based on a fork of Deutsche Postbank [pbrisk]
# Version:  0.5, copyright Wednesday, 18 September 2019
# Website:  https://github.com/sonntagsgesicht/businessdate
# License:  Apache License 2.0 (see LICENSE file)


from .businessperiod import BusinessPeriod
from .businessdate import BusinessDate
from .businessrange import BusinessRange


class BusinessSchedule(BusinessRange):
    def __init__(self, start, end, step, roll=None):
        """ class to build date schedules incl start and end date

        :param BusinessDate start: start date of schedule
        :param BusinessDate end: end date of schedule
        :param BusinessPeriod step: period distance of two dates
        :param BusinessDate roll: origin of schedule

        convenient class to build date schedules
        a schedule includes always start and end date
        and rolls on roll, i.e. builds a sequence by
        adding and/or substracting step to/from roll.
        start and end slice the relevant dates.
        """
        roll = roll if roll else end
        start, end = list(map(BusinessDate, (start, end)))
        super(BusinessSchedule, self).__init__(start, end, step, roll)
        if start not in self:
            self.insert(0, start)
        if end not in self:
            self.append(end)

    def first_stub_long(self):
        """ adjusts the schedule to have a long stub at the beginning,
            i.e. first period is longer a regular step.
        """
        if len(self) > 2:
            self.pop(1)
        return self

    def last_stub_long(self):
        """ adjusts the schedule to have a long stub at the end,
            i.e. last period is longer a regular step.
        """
        if len(self) > 2:
            self.pop(-2)
        return self



# -*- coding: utf-8 -*-

# businessdate
# ------------
# Python library for generating business dates for fast date operations
# and rich functionality.
# 
# Author:   sonntagsgesicht, based on a fork of Deutsche Postbank [pbrisk]
# Version:  0.5, copyright Wednesday, 18 September 2019
# Website:  https://github.com/sonntagsgesicht/businessdate
# License:  Apache License 2.0 (see LICENSE file)


import codecs

try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup

pkg = __import__('businessdate')

setup(
    name=pkg.__name__,
    description=pkg.__doc__,
    version=pkg.__version__,
    author=pkg.__author__,
    author_email=pkg.__email__,
    url=pkg.__url__,
    license=pkg.__license__,
    packages=(pkg.__name__,),
    package_data={pkg.__name__: list(pkg.__data__)},
    scripts=pkg.__scripts__,
    install_requires=pkg.__dependencies__,
    dependency_links=pkg.__dependency_links__,
    long_description='\n'+codecs.open('README.rst', encoding='utf-8').read(),
    platforms='any',
    classifiers=[
        'Development Status :: ' + pkg.__dev_status__,
        'Intended Audience :: Education',
        'Intended Audience :: Developers',
        'Intended Audience :: Financial and Insurance Industry',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Education',
        'Topic :: Office/Business',
        'Topic :: Office/Business :: Financial',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Utilities',
        'Topic :: Office/Business :: Scheduling',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)

